/**
  An extremely tiny and ill-featured camera class.
  Idea is you:
  
     call oglCameraLookAt(...) inside your display function, after gluPerspective.
     call oglCameraInit() inside your main function
  
  This file doesn't depend on anything else except glut and vec3.
*/
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include "osl/vec4.h" /* 3D vector class */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z /* convert a vec3 to three float arguments */


/* 3D orientation */
/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,0,1), z(0,-1,0) {} /* point down Y axis */
	
	/* Reorient this coordinate frame by this far along in the X and Y axes.  
	   "dx" is the distance along the z axis to push the x axis; 
	   "dy" the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct an orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};
ortho_frame camera_orient; /* camera orientation */
vec3 camera(0,0,0); /* position of camera (set up in main) */
bool key_down[256]; /* if true, the corresponding key is down */

/**
  Camera moves at the given real velocity in world-units per second.
*/
void oglCameraLookAt(float velocity=3.0)
{
	static float display_last_t=0.0;
	float display_cur_t=0.001*glutGet(GLUT_ELAPSED_TIME);
	float display_dt=display_cur_t-display_last_t; /* seconds per frame */
	display_last_t=display_cur_t;
	
	/* Physics and display timesteps: */
	float vel=velocity*display_dt; /* meters camera motion per frame */
	if (key_down[(int)'w']) camera-=vel*camera_orient.z;
	if (key_down[(int)'s']) camera+=vel*camera_orient.z;
	if (key_down[(int)'a']) camera-=vel*camera_orient.x;
	if (key_down[(int)'d']) camera+=vel*camera_orient.x;
	
	/* Don't allow head tilting! */
	camera_orient.x.z=0.0; 
	camera_orient.y=cross(camera_orient.z,camera_orient.x);
	camera_orient.orthonormalize();
	
	vec3 target=camera-camera_orient.z; /* camera looks down *negative* Z */
	gluLookAt(VEC3_TO_XYZ(camera), 
		VEC3_TO_XYZ(target), 
		VEC3_TO_XYZ(camera_orient.y)
	);
}

int mouse_x=0,mouse_y=0;
void camera_mouse(int button,int state,int x,int y) 
{ /* mouse being pressed or released--save position for motion */
	mouse_x=x; mouse_y=y;
}
void camera_motion(int x, int y) { 
	float scale=0.01; /* radians rotation/frame per mouse pixel motion */
	camera_orient.nudge(
		-scale*(x-mouse_x), 
		scale*(y-mouse_y)
	);
	mouse_x=x; mouse_y=y; /* save old mouse positions */
	glutPostRedisplay();
}
void camera_keyboard(unsigned char key,int x,int y) {
	key_down[key]=true;
	glutPostRedisplay();
}
void camera_keyboardUp(unsigned char key,int x,int y) {
	key_down[key]=false;
}

void oglCameraInit(void)
{
	glutMouseFunc(camera_mouse);
	glutMotionFunc(camera_motion); 
	glutKeyboardFunc(camera_keyboard);
	glutKeyboardUpFunc(camera_keyboardUp);
}

void oglCameraAxes(void)
{ /* Draw 3D axes on the ground */
	for (int axis=0;axis<3;axis++) {
		vec3 color(0.0);
		color[axis]=1.0;
		glColor3fv(color);
		
		glBegin(GL_LINES);
		glVertex3f(0,0,0);
		vec3 end(0.0); end[axis]=1.0;
		glVertex3fv(end);
		glEnd();
		
		glRasterPos3fv(end);
		glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,'X'+axis);
	}
}
